"""
This file contains biorbd specific functions for musculoskeletal analysis such as inverse or direct kinematics.
"""
try:
    import biorbd

    biordb_package = True
except ModuleNotFoundError:
    biordb_package = False
import numpy as np
from ..enums import InverseKinematicsMethods
from typing import Union
import time


class MskFunctions:
    def __init__(self, model: str, data_buffer_size: int = 1):
        """
        The MskFunctions contains all function for some musculoskeletal methods.

        Parameters
        ----------
        model : str
            Path to the biorbd model used to compute the kinematics.
        data_buffer_size: int
            The size of the buffer used to store the data.
        """
        if not biordb_package:
            raise ModuleNotFoundError(
                "Biorbd is not installed."
                " Please install it via"
                " 'conda install biorbd -cconda-forge' to use this function."
            )
        if isinstance(model, str):
            self.model = biorbd.Model(model)
        else:
            self.model = model
        self.process_time = []
        self.markers_buffer = []
        self.kin_buffer = []
        self.data_windows = data_buffer_size
        self.kalman = None

    def compute_inverse_kinematics(
        self,
        markers: np.ndarray,
        method: Union[InverseKinematicsMethods, str] = InverseKinematicsMethods.BiorbdLeastSquare,
        kalman_freq: Union[int, float] = 100,
        kalman: callable = None,
        custom_function: callable = None,
        **kwargs,
    ) -> tuple:
        """
        Function to apply the inverse kinematics using the markers data and a biorbd model type.

        Parameters
        ----------
        markers : numpy.array
            The experimental markers.
        kalman : biorbd.KalmanReconsMarkers
            The Kalman filter to use.
        kalman_freq : int
            The frequency of the Kalman filter.
        method : Union[InverseKinematicsMethods, str]
            The method to use to compute the inverse kinematics.
        custom_function : callable
            Custom function to use.

        Returns
        -------
        tuple
            The joint angle and velocity.
        """
        tic = time.time()
        if isinstance(method, str):
            if method in [t.value for t in InverseKinematicsMethods]:
                method = InverseKinematicsMethods(method)
            else:
                raise ValueError(f"Method {method} is not supported")

        if method == InverseKinematicsMethods.BiorbdKalman:
            self.kalman = kalman if kalman else self.kalman
            if not kalman and not self.kalman:
                freq = kalman_freq  # Hz
                params = biorbd.KalmanParam(freq)
                self.kalman = biorbd.KalmanReconsMarkers(self.model, params)
            markers_over_frames = []
            q = biorbd.GeneralizedCoordinates(self.model)
            q_dot = biorbd.GeneralizedVelocity(self.model)
            qd_dot = biorbd.GeneralizedAcceleration(self.model)
            for i in range(markers.shape[2]):
                markers_over_frames.append([biorbd.NodeSegment(m) for m in markers[:, :, i].T])

            q_recons = np.zeros((self.model.nbQ(), len(markers_over_frames)))
            q_dot_recons = np.zeros((self.model.nbQ(), len(markers_over_frames)))

            for i, targetMarkers in enumerate(markers_over_frames):
                self.kalman.reconstructFrame(self.model, targetMarkers, q, q_dot, qd_dot)
                q_recons[:, i] = q.to_array()
                q_dot_recons[:, i] = q_dot.to_array()

        elif method == InverseKinematicsMethods.BiorbdLeastSquare:
            ik = biorbd.InverseKinematics(self.model, markers)
            ik.solve("only_lm")
            q_recons = ik.q
            q_dot_recons = np.array([0] * ik.nb_q)[:, np.newaxis]

        elif method == InverseKinematicsMethods.Custom:
            if not custom_function:
                raise ValueError("No custom function provided.")
            q_recons = custom_function(markers, **kwargs)
            q_dot_recons = np.zerros((q_recons.shape()))
        else:
            raise ValueError(f"Method {method} is not supported")
        if len(self.kin_buffer) == 0:
            self.kin_buffer = [q_recons, q_dot_recons]
        else:
            self.kin_buffer[0] = np.append(self.kin_buffer[0], q_recons, axis=1)
            self.kin_buffer[1] = np.append(self.kin_buffer[1], q_dot_recons, axis=1)
        for i in range(len(self.kin_buffer)):
            self.kin_buffer[i] = self.kin_buffer[i][:, -self.data_windows :]
        self.process_time.append(time.time() - tic)
        return self.kin_buffer[0], self.kin_buffer[1]

    def compute_direct_kinematics(self, states: np.ndarray) -> np.ndarray:
        """
        Compute the direct kinematics using the joint angle and a biorbd model type.

        Parameters
        ----------
        states : np.ndarray
            The states to compute the direct kinematics.

        Returns
        -------
        np.ndarray
            The markers.
        """
        tic = time.time()
        if not biordb_package:
            raise ModuleNotFoundError(
                "Biorbd is not installed."
                " Please install it via"
                " 'conda install biorbd -cconda-forge' to use this function."
            )
        if isinstance(states, list):
            states = np.array(states)
        if states.shape[0] != self.model.nbQ():
            raise ValueError(f"States must have {self.model.nbQ()} rows.")
        if len(states.shape) != 2:
            states = states[:, np.newaxis]

        markers = np.zeros((3, self.model.nbMarkers(), states.shape[1]))
        for i in range(states.shape[1]):
            markers[:, :, i] = np.array([mark.to_array() for mark in self.model.markers(states[:, i])]).T
        if len(self.markers_buffer) == 0:
            self.markers_buffer = markers
        else:
            self.markers_buffer = np.append(self.markers_buffer, markers, axis=2)
        self.markers_buffer = self.markers_buffer[:, :, -self.data_windows :]
        self.process_time.append(time.time() - tic)
        return self.markers_buffer

    def get_mean_process_time(self):
        """
        Get the mean process time.

        Returns
        -------
        float
            The mean process time.
        """
        return np.mean(self.process_time)



"""
This file contains the Parameter class that define the device and markers classes.
"""
from math import ceil
from ..enums import DeviceType, MarkerType, InverseKinematicsMethods, RealTimeProcessingMethod, OfflineProcessingMethod
from ..processing.data_processing import RealTimeProcessing, OfflineProcessing, GenericProcessing
from ..processing.msk_functions import MskFunctions
from typing import Union
import numpy as np


class Param:
    def __init__(
        self,
        nb_channels: int,
        name: str = None,
        rate: float = None,
        system_rate: float = 100,
        data_window: int = None,
    ):
        """
        initialize the parameter class

        Parameters
        ----------
        nb_channels : int
            number of channels of the parameter
        name : str
            name of the parameter
        rate : float
            rate of the parameter
        system_rate : float
            rate of the system
        data_window : int
            size of the data window
        """
        self.nb_channels = nb_channels
        self.name = name
        self.rate = rate
        self.system_rate = system_rate
        self.sample = ceil(rate / self.system_rate)
        self.range = None
        self.raw_data = []
        self.data_window = data_window if data_window else int(rate)
        self.new_data = None

    def append_data(self, new_data: np.ndarray):
        """
        Append new data to the parameter to the raw data

        Parameters
        ----------
        new_data: np.ndarray
            new data to append to the buffer
        """
        if len(self.raw_data) == 0:
            self.raw_data = new_data
        elif self.raw_data.shape[-1] < self.data_window:
            self.raw_data = np.append(self.raw_data, new_data, axis=-1)
        else:
            self.raw_data = np.append(self.raw_data[..., new_data.shape[-1] :], new_data, axis=-1)


class Device(Param):
    """
    This class is used to store the available devices.
    """

    def __init__(
        self,
        device_type: DeviceType = DeviceType.Emg,
        nb_channels: int = 1,
        name: str = None,
        rate: float = 2000,
        system_rate: float = 100,
        channel_names: Union[list, str] = None,
    ):
        """
        Initialize the device class. A device is an electronic device that can be used to measure a parameter
         (e.g. EMG, treadmill, etc.).

        Parameters
        ----------
        device_type: DeviceType
            Type of the device.
        nb_channels: int
            Number of channels of the device.
        name: str
            Name of the device.
        rate: float
            Rate of the device.
        system_rate: float
            Rate of the system.
        channel_names: list
            Name of the channels of the device.
        """
        super().__init__(nb_channels, name, rate, system_rate)
        if isinstance(channel_names, str):
            channel_names = [channel_names]
        if channel_names:
            if nb_channels != len(channel_names):
                raise ValueError("The number of channels is not equal to the number of channel names.")
        else:
            channel_names = []
        self.device_range = None
        self.infos = None
        self.channel_names = channel_names
        self.interface = None
        self.device_type = device_type
        self.processed_data = None
        self.processing_method = None
        self.processing_function = None
        self.processing_method_kwargs = {}
        self.processing_method_changed = False
        self.processing_window = None

    def process(
        self,
        method: Union[str, RealTimeProcessingMethod, OfflineProcessingMethod] = None,
        custom_function: callable = None,
        **kwargs,
    ):
        """
        Process the data of the device. The raw data are stored in a buffer fill by the append data method.
        This method should be called after any get_data method, otherwise there will be not available data.

        Parameters
        ----------
        method: GenericProcessing
            Method to process the data.
        custom_function: callable
            Custom function to process the data.
        kwargs:
            Keyword arguments to pass to the method.
        """
        self.processing_method_kwargs.update(kwargs)
        if "processing_method" in self.processing_method_kwargs.keys():
            if method and method != self.processing_method_kwargs["processing_method"]:
                raise ValueError("You have enter two different type of method for the same function.")
            method = self.processing_method_kwargs["processing_method"]
        if (
            not method
            and not self.processing_method
            and "processing_method" not in self.processing_method_kwargs.keys()
        ):
            raise RuntimeError(
                "No method to process the data. Please specify a method with the argument 'processing_method'."
            )
        has_changed = self._check_if_has_changed(method, self.processing_method_kwargs)
        if "custom_function" in self.processing_method_kwargs.keys():
            custom_function = custom_function if custom_function else self.processing_method_kwargs["custom_function"]
            self.processing_method_kwargs.pop("custom_function")

        if not self.processing_function or has_changed:
            self._init_processing_method()

        if method == RealTimeProcessingMethod.Custom:
            if not custom_function:
                raise ValueError("No custom function to process the data.")
            self.processed_data = self.processing_function(
                custom_function, self.new_data, **self.processing_method_kwargs
            )
        else:
            self.processed_data = self.processing_function(self.new_data, **self.processing_method_kwargs)
        return self.processed_data

    def _init_processing_method(self):
        """
        Initialize the processing method.
        """
        self.processing_window = self.processing_window if self.processing_window else self.data_window
        if self.processing_method == RealTimeProcessingMethod.ProcessEmg:
            self.processing_function = RealTimeProcessing(self.rate, self.processing_window).process_emg
        elif self.processing_method == RealTimeProcessingMethod.ProcessGenericSignal:
            self.processing_function = RealTimeProcessing(self.rate, self.processing_window).process_generic_signal
        elif self.processing_method == RealTimeProcessingMethod.ProcessImu:
            self.processing_function = RealTimeProcessing(self.rate, self.processing_window).process_imu
        elif self.processing_method == RealTimeProcessingMethod.GetPeaks:
            self.processing_function = RealTimeProcessing(self.rate, self.processing_window).get_peaks
        elif self.processing_method == OfflineProcessingMethod.ProcessEmg:
            self.processing_function = OfflineProcessing(self.rate, self.processing_window).process_emg
        elif self.processing_method == OfflineProcessingMethod.ComputeMvc:
            self.processing_function = OfflineProcessing(self.rate, self.processing_window).compute_mvc
        elif (
            self.processing_method == RealTimeProcessingMethod.CalibrationMatrix
            or self.processing_method == OfflineProcessingMethod.CalibrationMatrix
        ):
            self.processing_function = GenericProcessing().calibration_matrix
        elif self.processing_method == RealTimeProcessingMethod.Custom:
            self.processing_function = RealTimeProcessing(self.rate, self.processing_window).custom_processing

    def _check_if_has_changed(self, method: GenericProcessing(), kwargs: dict) -> bool:
        """
        Check if the processing method has changed.

        Parameters
        ----------
        method: GenericProcessing
            Method to process the data.
        kwargs: dict
            Keyword arguments to pass to the method.

        Returns
        -------
        has_changed: bool
            True if the method has changed, False otherwise.

        """
        if isinstance(method, str):
            if method in [t.value for t in RealTimeProcessingMethod]:
                self.processing_method = RealTimeProcessingMethod(method)
            elif method not in [t.value for t in OfflineProcessingMethod]:
                self.processing_method = OfflineProcessingMethod(method)
            else:
                raise ValueError("The method is not valid.")
        has_changed = False
        if method and method != self.processing_method:
            has_changed = True
            self.processing_method = method

        if "processing_window" in kwargs:
            if kwargs["processing_window"] > self.data_window:
                raise ValueError("The processing windows is higher than the data buffer windows.")
            if kwargs["processing_window"] != self.processing_window:
                has_changed = True
                self.processing_window = kwargs["processing_window"]
            self.processing_method_kwargs.pop("processing_window")

        if self.new_data is None:
            raise RuntimeError("No data to process. Please run first the function get_device_data.")
        return has_changed


class MarkerSet(Param):
    """
    This class is used to store the available markers.
    """

    def __init__(
        self,
        nb_channels: int = 1,
        name: str = None,
        marker_names: Union[str, list] = None,
        rate: float = None,
        unlabeled: bool = False,
        system_rate: float = 100,
        unit: str = "m",
    ):
        """
        Initialize a marker set.

        Parameters
        ----------
        nb_channels: int
            Number of channels of the marker set (number of markers).
        name: str
            Name of the marker set.
        marker_names: str or list
            Name of the markers.
        rate: float
            Rate of the marker set.
        unlabeled: bool
            True if the marker set is unlabeled, False otherwise.
        system_rate: float
            Rate of the system.
        unit: str
            Unit of the marker set.
        """
        marker_type = MarkerType.Unlabeled if unlabeled else MarkerType.Labeled
        super(MarkerSet, self).__init__(nb_channels, name, rate, system_rate)
        if isinstance(marker_names, str):
            marker_names = [marker_names]
        if marker_names:
            if nb_channels != len(marker_names):
                raise ValueError("The number of channels and the number of markers names are not the same.")
        self.marker_names = marker_names
        self.subject_name = None
        self.interface = None
        self.marker_type = marker_type
        self.kin_data = None
        self.kin_method = None
        self.kin_method_kwargs = {}
        self.biorbd_model_path = None
        self.kalman = None
        self.msk_class = None
        self.unit = unit

    def get_kinematics(
        self,
        model_path: str = None,
        method: Union[InverseKinematicsMethods, str] = None,
        custom_function: callable = None,
        kin_data_window: int = None,
        **kwargs,
    ) -> tuple:
        """
        Function to apply the Kalman filter to the markers.
        The raw data are stored in a buffer fill by the append data method.
        This method should be called after any get_data method, otherwise there will be not available data.

        Parameters
        ----------
        model_path : str
            The biomod model used to compute the kinematics.
        method : Union[InverseKinematicsMethods, str]
            The method to use to compute the inverse kinematics.
        custom_function : callable
            Custom function to use.
        kin_data_window : int
            The size of the window to use to compute the kinematics.
        **kwargs : dict
            Keyword arguments to pass to the method.

        Returns
        -------
        tuple
            The joint angle and velocity.
        """
        if len(self.raw_data) == 0:
            raise RuntimeError(
                "No markers data to compute the kinematics." " Please run first the function get_markers_data."
            )
        method = method if method else self.kin_method
        self.kin_method_kwargs.update(kwargs)
        if "custom_function" in self.kin_method_kwargs.keys():
            custom_function = custom_function if custom_function else self.kin_method_kwargs["custom_function"]
        if self.new_data is None:
            raise RuntimeError("No data to process. Please run first the function get_markers_data.")
        if not method:
            raise RuntimeError("No method to compute the kinematics. Please specify a method.")
        if isinstance(method, str):
            if method in [t.value for t in InverseKinematicsMethods]:
                method = InverseKinematicsMethods(method)
            else:
                raise ValueError("The method is not valid.")
        kin_data_window = kin_data_window if kin_data_window else self.data_window
        if "model_path" in self.kin_method_kwargs.keys():
            model_path = self.kin_method_kwargs["model_path"]
        model_path = model_path if model_path else self.biorbd_model_path
        if model_path is None and not "model_path" in self.kin_method_kwargs.keys():
            raise ValueError("No model to compute the kinematics.")
        if not self.msk_class:
            self.msk_class = MskFunctions(model_path, kin_data_window)
        if method == InverseKinematicsMethods.Custom:
            if not custom_function:
                raise ValueError("No custom function to process the data.")
            self.kin_data = self.msk_class.compute_inverse_kinematics(
                self.new_data, method, self.rate, custom_function=custom_function, **self.kin_method_kwargs
            )
        else:
            self.kin_data = self.msk_class.compute_inverse_kinematics(
                self.new_data, method, self.rate, **self.kin_method_kwargs
            )
        return self.kin_data



"""
This example shows how to save data in a pickle file (with a *.bio extension) using the pre-implemented function. The special feature of this function is that the data is added to the file without reading the whole file, so it is a fast way to save data to a binary file inside a loop.
So it's a quick way to save data to a binary file inside a loop. You might want to do this for the online data stream in case of a malfunction and to not fill a table with too much data.
The file can then be read using the load function which concatenates all the data in the file and returns a dictionary containing all the data.
Keep in mind that the load function has to read each row one after the other, so if you have a lot of rows, it may take a while.
The load function takes the number of lines you want to read as an argument in case it opens too slowly.
"""

import os
from biosiglive import save, load
import numpy as np

if __name__ == "__main__":
    data = {
        "data_np": np.random.rand(2, 20),
        "data_list": [5, 8],
        "data_dict": {"data_np": np.random.rand(2, 20)},
        "data_int": 1,
    }
    i = 0
    while i != 50:
        save(data, "test")
        i += 1
    data_loaded = load("test.bio")
    os.remove("test.bio")



"""
This file contains a class that allows to stream data from a source and start some multiprocess
 to process and disseminate. It is a work in progress so basic functions are available but need to be fine-tuned.
"""

from typing import Union
from time import time, sleep, strftime
import datetime
import numpy as np
import multiprocessing as mp
from biosiglive.streaming.server import Server
from ..file_io.save_and_load import save
from ..interfaces.generic_interface import GenericInterface
from ..interfaces.param import Device, MarkerSet
from ..gui.plot import LivePlot
from .utils import dic_merger


# TODO add enum for command type
class StreamData:
    def __init__(self, stream_rate: int = 100):
        """
        Initialize the StreamData class.
        Careful this class do not return anything, you will have to turn the save option to True to save the data.

        Parameters
        ----------
        stream_rate: int
            The stream rate of the data.
        """
        self.process = mp.Process
        self.devices = []
        self.marker_sets = []
        self.plots = []
        self.stream_rate = stream_rate
        self.interfaces_type = []
        self.processes = []
        self.interfaces = []
        self.multiprocess_started = False

        # Multiprocessing stuff
        manager = mp.Manager()
        self.queue = manager.Queue
        self.event = manager.Event
        self.device_queue_in = []
        self.device_queue_out = []
        self.kin_queue_in = []
        self.kin_queue_out = []
        self.plots_queue = []
        self.device_event = []
        self.is_device_data = self.event()
        self.is_kin_data = self.event()
        self.interface_event = []
        self.kin_event = []
        self.custom_processes = []
        self.custom_processes_kwargs = []
        self.custom_processes_names = []
        self.custom_queue_in = []
        self.custom_queue_out = []
        self.custom_event = []
        self.save_data = None
        self.save_path = None
        self.save_frequency = None
        self.plots_multiprocess = False
        self.device_buffer_size = []
        self.marker_set_buffer_size = []
        self.raw_plot = None
        self.data_to_plot = None

        # Server stuff
        self.start_server = None
        self.server_ip = None
        self.ports = []
        self.client_type = None
        self.count_server = 0
        self.server_queue = []
        self.device_decimals = 8
        self.kin_decimals = 6

    def _add_device(self, device: Device):
        """
        Add a device to the stream.

        Parameters
        ----------
        device: Device
            Device to add.
        """
        self.devices.append(device)
        self.device_queue_in.append(self.queue())
        self.device_queue_out.append(self.queue())
        self.device_event.append(self.event())

    def add_interface(self, interface: GenericInterface()):
        """
        Add an interface to the stream.

        Parameters
        ----------
        interface: GenericInterface
            Interface to add. Interface should inherit from the generic interface.
        """
        if self.multiprocess_started:
            raise Exception("Cannot add interface after the stream has started.")
        self.interfaces.append(interface)
        self.interfaces_type.append(interface.interface_type)
        self.interface_event.append(self.event())
        for device in interface.devices:
            self._add_device(device)
        for marker in interface.marker_sets:
            self._add_marker_set(marker)
        if len(self.interfaces) > 1:
            raise ValueError("Only one interface can be added for now.")

    def add_server(
        self,
        server_ip: str = "127.0.0.1",
        ports: Union[int, list] = 50000,
        client_type: str = "TCP",
        device_buffer_size: Union[int, list] = None,
        marker_set_buffer_size: [int, list] = None,
    ):
        """
        Add a server to the stream.

        Parameters
        ----------
        server_ip: str
            The ip address of the server.
        ports: int or list
            The port(s) of the server.
        client_type: str
            The type of client to use. Can be TCP.
        device_buffer_size: int or list
            The size of the buffer for the devices.
        marker_set_buffer_size: int or list
            The size of the buffer for the marker sets.
        """
        if self.multiprocess_started:
            raise Exception("Cannot add interface after the stream has started.")
        self.server_ip = server_ip
        self.ports = ports
        if not isinstance(self.ports, list):
            self.ports = [self.ports]

        for p in range(len(self.ports)):
            self.server_queue.append(self.queue())
        self.client_type = client_type

        if not device_buffer_size:
            device_buffer_size = [None] * len(self.devices)
        if isinstance(device_buffer_size, list):
            if len(device_buffer_size) != len(self.devices):
                raise ValueError("The device buffer size list should have the same length as the number of devices.")
            self.device_buffer_size = device_buffer_size
        elif isinstance(device_buffer_size, int):
            self.device_buffer_size = [device_buffer_size] * len(self.devices)

        if not marker_set_buffer_size:
            marker_set_buffer_size = [None] * len(self.marker_sets)
        if isinstance(marker_set_buffer_size, list):
            if len(marker_set_buffer_size) != len(self.marker_sets):
                raise ValueError(
                    "The marker set buffer size list should have the same length as the number of marker sets."
                )
            self.marker_set_buffer_size = marker_set_buffer_size
        elif isinstance(marker_set_buffer_size, int):
            self.marker_set_buffer_size = [marker_set_buffer_size] * len(self.marker_sets)

    def start(self, save_streamed_data: bool = False, save_path: str = None, save_frequency: int = None):
        """
        Start the stream.

        Parameters
        ----------
        save_streamed_data: bool
            If True, the streamed data will be saved.
        save_path: str
            The path to save the streamed data.
        save_frequency:
            The frequency at which the data will be saved.
        """
        self.save_data = save_streamed_data
        self.save_path = save_path if save_path else f"streamed_data_{strftime('%Y%m%d_%H%M%S')}.bio"
        self.save_frequency = save_frequency if save_frequency else self.stream_rate
        self._init_multiprocessing()

    def _add_marker_set(self, marker: MarkerSet):
        """
        Add a marker set to the stream.

        Parameters
        ----------
        marker: MarkerSet
            Marker set to add from given interface.
        """
        self.marker_sets.append(marker)
        self.kin_queue_in.append(self.queue())
        self.kin_queue_out.append(self.queue())
        self.kin_event.append(self.event())

    # TODO : add buffer directly in the server
    def device_processing(self, device_idx: int):
        """
        Process the data from the device

        Parameters
        ----------
        device_idx: int
            The index of the device in the list of devices.
        """
        if self.device_buffer_size:
            if not self.device_buffer_size[device_idx]:
                self.device_buffer_size[device_idx] = self.devices[device_idx].rate
            buffer_size = self.device_buffer_size[device_idx]
        else:
            buffer_size = self.devices[device_idx].rate
        device_data = []
        while True:
            self.is_device_data.wait()
            try:
                device_data = self.device_queue_in[device_idx].get_nowait()
                is_working = True
            except Exception:
                is_working = False
            if is_working:
                self.devices[device_idx].new_data = device_data
                self.devices[device_idx].append_data(device_data)
                processed_data = self.devices[device_idx].process(**self.devices[device_idx].processing_method_kwargs)
                self.device_queue_out[device_idx].put_nowait({"processed_data": processed_data[:, -buffer_size:]})
                self.device_event[device_idx].set()

    def recons_kin(self, marker_set_idx: int):
        """
        Compute inverse kinematics from markers.

        Parameters
        ----------
        marker_set_idx: int
            Index of the marker set in the list of markers.
        """
        if self.marker_set_buffer_size:
            if not self.marker_set_buffer_size[marker_set_idx]:
                self.marker_set_buffer_size[marker_set_idx] = self.marker_sets[marker_set_idx].rate
            buffer_size = self.marker_set_buffer_size[marker_set_idx]
        else:
            buffer_size = self.marker_sets[marker_set_idx].rate
        if "model_path" not in self.marker_sets[marker_set_idx].kin_method_kwargs.keys():
            raise ValueError("No model to compute the kinematics.")
        markers = []
        while True:
            self.is_kin_data.wait()
            try:
                markers = self.kin_queue_in[marker_set_idx].get_nowait()
                is_working = True
            except Exception:
                is_working = False
            if is_working:
                self.marker_sets[marker_set_idx].new_data = markers
                self.marker_sets[marker_set_idx].append_data(markers)
                states, _ = self.marker_sets[marker_set_idx].get_kinematics(
                    **self.marker_sets[marker_set_idx].kin_method_kwargs
                )
                self.kin_queue_out[marker_set_idx].put_nowait({"kinematics_data": states[:, -buffer_size:]})
                self.kin_event[marker_set_idx].set()

    def open_server(self, server_idx: int):
        """
        Open the server to send data from the devices.
        """
        server = Server(self.server_ip, self.ports[server_idx], server_type=self.client_type)
        server.start()
        while True:
            connection, message = server.client_listening()
            data_queue = []
            while len(data_queue) == 0:
                # use Try statement as the queue can be empty and is_empty function is not reliable.
                try:
                    data_queue = self.server_queue[server_idx].get_nowait()
                    is_working = True
                except Exception:
                    is_working = False
                if is_working:  # use this method to avoid blocking the server with Windows os.
                    server.send_data(data_queue, connection, message)

    def _init_multiprocessing(self):
        """
        Initialize the multiprocessing.
        """
        processes = []
        for i in range(len(self.interfaces)):
            processes.append(
                self.process(name="reader", target=StreamData.save_streamed_data, args=(self, i), daemon=True)
            )

        for d, device in enumerate(self.devices):
            if device.processing_method is not None:
                processes.append(
                    self.process(
                        name=f"process_{device.name}",
                        target=StreamData.device_processing,
                        args=(
                            self,
                            d,
                        ),
                        daemon=True,
                    )
                )
        for i in range(len(self.ports)):
            processes.append(
                self.process(name="listen" + f"_{i}", target=StreamData.open_server, args=(self, i), daemon=True)
            )

        for p, plot in enumerate(self.plots):
            for device in self.devices:
                for marker_set in self.marker_sets:
                    if self.data_to_plot[p] not in device.name and self.data_to_plot[p] not in marker_set.name:
                        raise ValueError(f"The name of the data to plot ({self.data_to_plot[p]}) is not correct.")
            if self.plots_multiprocess:
                processes.append(self.process(name="plot", target=StreamData.plot_update, args=(self, p), daemon=True))
            else:
                processes.append(self.process(name="plot", target=StreamData.plot_update, args=(self, -1), daemon=True))
                break

        for m, marker in enumerate(self.marker_sets):
            if marker.kin_method:
                processes.append(
                    self.process(
                        name=f"process_{marker.name}",
                        target=StreamData.recons_kin,
                        args=(
                            self,
                            m,
                        ),
                        daemon=True,
                    )
                )

        for i, funct in enumerate(self.custom_processes):
            processes.append(
                self.process(
                    name=self.custom_processes_names[i],
                    target=funct,
                    args=(self,),
                    kwargs=self.custom_processes_kwargs[i],
                    daemon=True,
                )
            )
        for p in processes:
            p.start()
        self.multiprocess_started = True
        for p in processes:
            p.join()

    def _check_nb_processes(self):
        """
        compute the number of process.
        """
        nb_processes = 0
        for device in self.devices:
            if device.process_method is not None:
                nb_processes += 1
        if self.start_server:
            nb_processes += len(self.ports)
        nb_processes += len(self.plots)
        nb_processes += len(self.interfaces)
        for marker in self.marker_sets:
            if marker.kin_method:
                nb_processes += 1
        nb_processes += len(self.custom_processes)
        return nb_processes

    def add_plot(
        self,
        plot: Union[LivePlot, list],
        data_to_plot: Union[str, list],
        raw: Union[bool, list] = None,
        multiprocess=False,
    ):
        """
        Add a plot to the live data. Still Not working for now.

        Parameters
        ----------
        plot: Union[LivePlot, list]
            Plot to add.
        data_to_plot: Union[str, list]
            Name of the data to plot.
        raw: Union[bool, list]
            If True, the raw data will be plotted.
        multiprocess: bool
            If True, if several plot each plot will be on a separate process. If False, each plot will be on the same one.
        """
        raise NotImplementedError("Plot are not implemented yet with StreamData class.")
        # if isinstance(data_to_plot, str):
        #     data_to_plot = [data_to_plot]
        # if isinstance(raw, bool):
        #     raw = [raw]
        # if len(data_to_plot) != len(raw):
        #     raise ValueError("The length of the data to plot and the raw list must be the same.")
        # if not raw:
        #     raw = [True] * len(data_to_plot)
        # self.plots_queue.append(self.queue())
        # self.raw_plot = raw
        # self.data_to_plot = data_to_plot
        # if self.multiprocess_started:
        #     raise Exception("Cannot add plot after the stream has started.")
        # self.plots_multiprocess = multiprocess
        # if not isinstance(plot, list):
        #     plot = [plot]
        # for plt in plot:
        #     if plt.rate:
        #         if plt.rate > self.stream_rate:
        #             raise ValueError("Plot rate cannot be higher than stream rate.")
        #     self.plots.append(plt)

    def plot_update(self, plot_idx: int = -1):
        """
        Update the plots.

        Parameters
        ----------
        plot_idx: int
            index of the plot to update. If -1, all plots will be updated.
        """
        if plot_idx == -1:
            plots = self.plots
            queue = self.plots_queue[0]
        else:
            plots = self.plots[plot_idx]
            queue = self.plots_queue[plot_idx]
        data_to_plot = []
        data = None
        device_names = []
        marker_set_names = []
        for device in self.devices:
            device_names.append(device.name)
        for marker in self.marker_sets:
            marker_set_names.append(marker.name)
        while True:
            try:
                data = queue.get_nowait()
                is_working = True
            except Exception:
                is_working = False
            if is_working:
                for p, plot in enumerate(plots):
                    if self.data_to_plot[p] in device_names:
                        if not self.raw_plot[p]:
                            data_to_plot = data["proc_device_data"][device_names.index(self.data_to_plot[p])]
                        else:
                            data_to_plot = data["raw_device_data"][device_names.index(self.data_to_plot[p])]
                    if self.data_to_plot[p] in marker_set_names:
                        if not self.raw_plot[p]:
                            data_to_plot = data["kinematics_data"][marker_set_names.index(self.data_to_plot[p])]
                        else:
                            data_to_plot = data["marker_set_data"][marker_set_names.index(self.data_to_plot[p])][
                                :, :, -1
                            ].T
                    plot.update(data_to_plot)

    def save_streamed_data(self, interface_idx: int):
        """
        Stream, process and save the data.

        Parameters
        ----------
        interface_idx: idx
            Interface index to use from the interface list. for now only one interface is supported.

        """
        initial_time = 0
        iteration = 0
        dic_to_save = {}
        save_count = 0
        self.save_frequency = self.save_frequency if self.save_frequency else self.stream_rate
        interface = self.interfaces[interface_idx]
        saving_time = None
        while True:
            data_dic = {}
            proc_device_data = []
            raw_device_data = []
            raw_markers_data = []
            all_device_data = []
            all_markers_tmp = []
            kin_data = []
            tic = time()
            if iteration == 0:
                initial_time = time() - tic
            interface_latency = interface.get_latency()
            is_frame = interface.get_frame()
            absolute_time_frame = datetime.datetime.now()
            absolute_time_frame_dic = {
                "day": absolute_time_frame.day,
                "hour": absolute_time_frame.hour,
                "hour_s": absolute_time_frame.hour * 3600,
                "minute": absolute_time_frame.minute,
                "minute_s": absolute_time_frame.minute * 60,
                "second": absolute_time_frame.second,
                "millisecond": int(absolute_time_frame.microsecond / 1000),
                "millisecond_s": int(absolute_time_frame.microsecond / 1000) * 0.001,
            }
            self.is_kin_data.clear()
            self.is_device_data.clear()
            if is_frame:
                if iteration == 0:
                    print("Data start streaming")
                    iteration = 1
                if len(interface.devices) != 0:
                    all_device_data = interface.get_device_data(device_name="all", get_frame=False)
                    if not isinstance(all_device_data, list):
                        all_device_data = [all_device_data]
                    self.is_device_data.set()
                    for i in range(len(all_device_data)):
                        if self.devices[i].processing_method is not None:
                            self.device_queue_in[i].put_nowait(all_device_data[i])
                if len(interface.marker_sets) != 0:
                    all_markers_tmp, _ = interface.get_marker_set_data(get_frame=False)
                    if not isinstance(all_markers_tmp, list):
                        all_markers_tmp = [all_markers_tmp]
                    self.is_kin_data.set()
                    for i in range(len(self.marker_sets)):
                        if self.marker_sets[i].kin_method is not None:
                            self.kin_queue_in[i].put_nowait(all_markers_tmp[i])
                time_to_get_data = time() - tic
                tic_process = time()
                if len(interface.devices) != 0:
                    for i in range(len(interface.devices)):
                        if self.devices[i].processing_method is not None:
                            self.device_event[i].wait()
                            device_data = self.device_queue_out[i].get_nowait()
                            self.device_event[i].clear()
                            proc_device_data.append(
                                np.around(device_data["processed_data"], decimals=self.device_decimals)
                            )
                        raw_device_data.append(np.around(all_device_data[i], decimals=self.device_decimals))
                    data_dic["proc_device_data"] = proc_device_data
                    data_dic["raw_device_data"] = raw_device_data

                if len(interface.marker_sets) != 0:
                    for i in range(len(interface.marker_sets)):
                        if self.marker_sets[i].kin_method is not None:
                            self.kin_event[i].wait()
                            kin_data_proc = self.kin_queue_out[i].get_nowait()
                            self.kin_event[i].clear()
                            kin_data.append(np.around(kin_data_proc["kinematics_data"], decimals=self.kin_decimals))
                        raw_markers_data.append(np.around(all_markers_tmp[i], decimals=self.kin_decimals))
                    data_dic["kinematics_data"] = kin_data
                    data_dic["marker_set_data"] = raw_markers_data
                process_time = time() - tic_process  # time to process all data

                for i in range(len(self.ports)):
                    try:
                        self.server_queue[i].get_nowait()
                    except Exception:
                        pass
                    self.server_queue[i].put_nowait(data_dic)
                if len(self.plots) != 0:
                    size = 1 if not self.plots_multiprocess else len(self.plots)
                    for i in range(size):
                        try:
                            self.plots_queue[i].get_nowait()
                        except Exception:
                            pass
                        self.plots_queue[i].put_nowait(data_dic)

                data_dic["absolute_time_frame"] = absolute_time_frame_dic
                data_dic["interface_latency"] = interface_latency
                data_dic["process_time"] = process_time
                data_dic["initial_time"] = initial_time
                data_dic["time_to_get_data"] = time_to_get_data

                # Save data
                if self.save_data is True:
                    tic_save = time()
                    data_dic["saving_time"] = saving_time
                    dic_to_save = dic_merger(data_dic, dic_to_save)
                    if save_count == int(self.stream_rate / self.save_frequency):
                        save(data_dic, self.save_path)
                        dic_to_save = {}
                        save_count = 0
                    save_count += 1
                    saving_time = time() - tic_save

                if tic - time() < 1 / self.stream_rate:
                    sleep(1 / self.stream_rate - (tic - time()))
                else:
                    print(
                        f"WARNING: Stream rate ({self.stream_rate}) is too high for the computer."
                        f"The actual stream rate is {1 / (tic - time())}"
                    )

    def stop(self):
        """
        Stop the stream
        """
        for process in self.processes:
            process.terminate()
            process.join()
